#!/usr/bin/env ruby

#  Copyright (C) 2017 Registro.br. All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
# 1. Redistribution of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 
# THIS SOFTWARE IS PROVIDED BY REGISTRO.BR ``AS IS AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIE OF FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
# EVENT SHALL REGISTRO.BR BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
# OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
# TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
# DAMAGE.

# $Id$

require 'date'
require 'test/unit'

if RUBY_VERSION < "1.9"
  require 'isavail'
else
  require_relative 'isavail'
end

#############################################
##                Tests                    ##
#############################################

class TestResponse < Test::Unit::TestCase
  def test_query
    query = Query.new 'example.com.br'
    query.cookie = '8a6e52837fc2c68f7a4a'
    query.language = 'EN'

    assert_equal("1 8a6e52837fc2c68f7a4a 0 #{query.query_id} example.com.br 1", query.to_s)
  end

  def test_query_via_proxy
    query = Query.new 'example.com.br'
    query.cookie = '8a6e52837fc2c68f7a4a'
    query.language = 'EN'
    query.proxy = '200.160.2.3'

    assert_equal("[200.160.2.3] 1 8a6e52837fc2c68f7a4a 0 #{query.query_id} example.com.br 1", query.to_s)
  end

  def test_wrong_cookie
    data = "CK 8a6e52837fc2c68f7a4a 12345"
    
    response = Response.new data
    assert_equal(-1, response.status)
    assert_equal(12345, response.query_id)
    assert_equal('8a6e52837fc2c68f7a4a', response.cookie)
  end

  def test_available
    data = "% Copyright registro.br\r\n" +
      "ST 0 12345\r\n" +
      "example.com.br"

    response = Response.new data
    assert_equal(0, response.status)
    assert_equal(12345, response.query_id)
    assert_equal('example.com.br', response.fqdn)
  end

  def test_available_with_fqdn_ace
    data = "% Copyright registro.br\r\n" +
      "ST 0 12345\r\n" +
      "xn--example-3wa.com.br|exampleã.com.br"

    response = Response.new data
    assert_equal(0, response.status)
    assert_equal(12345, response.query_id)
    assert_equal('exampleã.com.br', response.fqdn)
    assert_equal('xn--example-3wa.com.br', response.fqdn_ace)
  end

  def test_available_with_tickets
    data = "% Copyright registro.br\r\n" +
      "ST 1 12345\r\n" +
      "example.com.br\r\n" +
      "2567849|2567856"

    response = Response.new data
    assert_equal(1, response.status)
    assert_equal(12345, response.query_id)
    assert_equal('example.com.br', response.fqdn)
    assert_equal(['2567849', '2567856'], response.tickets)
  end

  def test_registered
    data = "% Copyright registro.br\r\n" +
      "ST 2 12345\r\n" +
      "example.eng.br\r\n" +
      "2007-03-15|published|fork.example.eng.br|example.eng.br\r\n" +
      "blog|flog|sec3|vlog|wiki"

    response = Response.new data
    assert_equal(2, response.status)
    assert_equal(12345, response.query_id)
    assert_equal('example.eng.br', response.fqdn)
    assert_equal(Date.parse('2007-03-15'), response.expiration_date)
    assert_equal('published', response.publication)
    assert_equal(['fork.example.eng.br', 'example.eng.br'], response.nameservers)
    assert_equal(['blog', 'flog', 'sec3', 'vlog', 'wiki'], response.suggestions)
  end

  def test_unavailable
    data = "% Copyright registro.br\r\n" +
      "ST 3 12345\r\n" +
      "example.com.br\r\n" +
      "reserved CG\r\n" +
      "agr|art|esp|etc|far|imb|ind|inf|rec|srv|tmp|tur"

    response = Response.new data
    assert_equal(3, response.status)
    assert_equal(12345, response.query_id)
    assert_equal('example.com.br', response.fqdn)
    assert_equal('reserved CG', response.reason)
    assert_equal(['agr', 'art', 'esp', 'etc', 'far', 'imb', 'ind', 'inf', 'rec', 'srv', 'tmp', 'tur'], response.suggestions)
  end

  def test_invalid
    data = "% Copyright registro.br\r\n" +
      "ST 4 12345\r\n" +
      "example.com"

    response = Response.new data
    assert_equal(4, response.status)
    assert_equal(12345, response.query_id)
    assert_equal('example.com', response.fqdn)
    assert_equal('', response.reason)
  end

  def test_waiting
    data = "% Copyright registro.br\r\n" +
      "ST 5 12345\r\n" +
      "example.com.br"

    response = Response.new data
    assert_equal(5, response.status)
    assert_equal(12345, response.query_id)
    assert_equal('example.com.br', response.fqdn)
  end

  def test_release_process
    data = "% Copyright registro.br\r\n" +
      "ST 6 12345\r\n" +
      "example.com.br\r\n" +
      "2007-01-15 15:00:00|2007-01-30 15:00:00"

    response = Response.new data
    assert_equal(6, response.status)
    assert_equal(12345, response.query_id)
    assert_equal('example.com.br', response.fqdn)
    assert_equal(DateTime.parse_nicbr('2007-01-15 15:00:00'), response.release_process_begin)
    assert_equal(DateTime.parse_nicbr('2007-01-30 15:00:00'), response.release_process_end)
  end

  def test_release_process_with_tickets
    data = "% Copyright registro.br\r\n" +
      "ST 7 12345\r\n" +
      "example.com.br\r\n" +
      "2007-01-15 15:00:00|2007-01-30 15:00:00\r\n" +
      "2567849|2567856"

    response = Response.new data
    assert_equal(7, response.status)
    assert_equal(12345, response.query_id)
    assert_equal('example.com.br', response.fqdn)
    assert_equal(DateTime.parse_nicbr('2007-01-15 15:00:00'), response.release_process_begin)
    assert_equal(DateTime.parse_nicbr('2007-01-30 15:00:00'), response.release_process_end)
    assert_equal(['2567849', '2567856'], response.tickets)
  end

  def test_error
    data = "% Copyright registro.br\r\n" +
      "ST 8\r\n" +
      "Query rate limit exceeded"

    begin
      response = Response.new data
    rescue => e
      assert_equal('Query rate limit exceeded', e.message)
    end
  end
end
