#  Copyright (C) 2017 Registro.br. All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
# 1. Redistribution of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 
# THIS SOFTWARE IS PROVIDED BY REGISTRO.BR ``AS IS AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIE OF FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
# EVENT SHALL REGISTRO.BR BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
# OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
# TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
# DAMAGE.

# $Id$

require 'date'
require 'socket'

MAX_UDP_SIZE = 512

#############################################
##           DateTime extension            ##
#############################################

class DateTime
  def self.parse_nicbr(date)
    DateTime.strptime(date, '%Y-%m-%d %H:%M:%S')
  end
end

#############################################
##                Query                    ##
#############################################

class Query
  attr_writer :proxy, :cookie, :fqdn, :suggest
  attr_reader :query_id

  def initialize(fqdn)
    @proxy = ''
    @version = 2
    @cookie = '00000000000000000000'
    @language = 1
    @query_id = rand 2**32
    @fqdn = fqdn
    @suggest = true
  end

  def cookie_file=(value)
    begin
      @cookie = IO.read value
    rescue
      # File does not exist
    end
  end

  def language=(value)
    if value == 'PT'
      @language = 1
    elsif value == 'EN'
      @language = 0
    end
  end

  def options=(values)
    @proxy = values[:proxy] if values.include? :proxy
    self.cookie_file = values[:cookie_file] if values.include? :cookie_file and not values[:cookie_file].empty?
    self.language = values[:language] if values.include? :language
    @fqdn = values[:fqdn] if values.include? :fqdn
    @suggest = values[:suggest] if values.include? :suggest
  end

  def to_s
    "#{'[' + @proxy + '] ' if @proxy.size > 0}" +
      "#{@version} " +
      "#{@cookie} " +
      "#{@language} " +
      "#{@query_id} " +
      "#{@fqdn} " +
      "#{'1' if @suggest == true}#{'0' if @suggest == false}"
  end
end

#############################################
##                Response                 ##
#############################################

module Undefined
  def parse(data)
  end

  def status_to_s
    'Undefined'
  end

  def to_s
    ''
  end
end

module AvailableOrWaiting
  def parse(data)
    parse_fqdn @data_lines[2] if @data_lines.size > 2
  end

  def status_to_s
    if @status == 0
      'Available'
    
    elsif @status == 5
      'Release process waiting'
    end
  end

  def to_s
    "Query ID: #{@query_id}\n" +
      "Domain name: #{@fqdn}\n" +
      "Response status: #{@status} (#{status_to_s})"
  end
end

module AvailableWithTickets
  attr_reader :tickets

  def parse(data)
    parse_fqdn @data_lines[2] if @data_lines.size > 2

    @tickets = []
    @data_lines[3].split('|').each { |ticket| @tickets << ticket } if @data_lines.size > 3
  end

  def status_to_s
    'Available with active tickets'
  end

  def to_s
    "Query ID: #{@query_id}\n" +
      "Domain name: #{@fqdn}\n" +
      "Response status: #{@status} (#{status_to_s})\n" +
      "Tickets: ##{@tickets.join ', #'}"
  end
end

module Registered
  attr_reader :expiration_date, :publication, :nameservers, :suggestions

  def parse(data)
    parse_fqdn @data_lines[2] if @data_lines.size > 2

    @expiration_date = Date.new
    @publication = ''
    @nameservers = []
    if @data_lines.size > 3
      data_info = @data_lines[3].split '|'
      if data_info.size >= 3
        @expiration_date = Date.parse data_info[0]
        @publication = data_info[1]
      end

      2.upto(data_info.size-1) {|i| @nameservers << data_info[i]}
    end

    @suggestions = []
    @data_lines[4].split('|').each {|tld| @suggestions << tld} if @data_lines.size > 4
  end

  def status_to_s
    'Registered'
  end

  def to_s
    string = "Query ID: #{@query_id}\n" +
      "Domain name: #{@fqdn}\n" +
      "Response status: #{@status} (#{status_to_s})\n" +
      "Expiration date: #{@expiration_date.strftime('%Y-%m-%d')}\n" +
      "Publication Status: #{@publication}\n" +
      "Nameservers: #{@nameservers.join ', '}"
    string += "\nSuggestions: #{@suggestions.join ', '}" unless @suggestions.empty?
    string
  end
end

module Unavailable
  attr_reader :reason, :suggestions

  def parse(data)
    parse_fqdn @data_lines[2] if @data_lines.size > 2
    @reason = ''
    @reason = @data_lines[3] if @data_lines.size > 3
    @suggestions = []
    @data_lines[4].split('|').each {|tld| @suggestions << tld} if @data_lines.size > 4
  end

  def status_to_s
    'Unavailable'
  end

  def to_s
    string = "Query ID: #{@query_id}\n" +
      "Domain name: #{@fqdn}\n" +
      "Response status: #{@status} (#{status_to_s})\n" +
      "Reason: #{@reason}"
    string += "\nSuggestions: #{@suggestions.join ', '}" unless @suggestions.empty?
    string
  end
end

module Invalid
  attr_reader :reason

  def parse(data)
    parse_fqdn @data_lines[2] if @data_lines.size > 2
    @reason = ''
    @reason = @data_lines[3] if @data_lines.size > 3
  end

  def status_to_s
    'Invalid query'
  end

  def to_s
    "Query ID: #{@query_id}\n" +
      "Domain name: #{@fqdn}\n" +
      "Response status: #{@status} (#{status_to_s})\n" +
      "Reason: #{@reason}"
  end
end

module ReleaseInProgress
  attr_reader :release_process_begin, :release_process_end, :tickets

  def parse(data)
    parse_fqdn @data_lines[2] if @data_lines.size > 2

    @release_process_begin = DateTime.new
    @release_process_end = DateTime.new
    if @data_lines.size > 3
      data_info = @data_lines[3].split '|'
      if data_info.size >= 2
        @release_process_begin, @release_process_end =
          DateTime.parse_nicbr(data_info[0]), DateTime.parse_nicbr(data_info[1])
      end
    end

    @tickets = []
    @data_lines[4].split('|').each { |ticket| @tickets << ticket } if @data_lines.size > 4
  end

  def status_to_s
    if @status == 6
      'Release process in progress'
    
    elsif @status == 7
      'Release process in progress with active tickets'
    end
  end

  def to_s
    string = "Query ID: #{@query_id}\n" +
      "Domain name: #{@fqdn}\n" +
      "Response status: #{@status} (#{status_to_s})\n" +
      "Release Process:\n" +
      "  Start date: #{@release_process_begin.strftime('%Y-%m-%d %H:%M')}\n" +
      "  End date: #{@release_process_end.strftime('%Y-%m-%d %H:%M')}"
    string += "\nTickets: ##{@tickets.join ', #'}" unless @tickets.empty?
    string
  end
end

module CompetitiveReleaseProcess
  attr_reader :release_process_begin, :release_process_end, :accepting_new_tickets_until, :tickets

  def parse(data)
    parse_fqdn @data_lines[2] if @data_lines.size > 2

    @release_process_begin = DateTime.new
    @release_process_end = DateTime.new
    if @data_lines.size > 3
      data_info = @data_lines[3].split '|'
      if data_info.size >= 2
        @release_process_begin, @release_process_end, @accepting_new_tickets_until =
          DateTime.parse_nicbr(data_info[0]), DateTime.parse_nicbr(data_info[1]), DateTime.parse_nicbr(data_info[2])
      end
    end

    @tickets = []
    @data_lines[4].split('|').each { |ticket| @tickets << ticket } if @data_lines.size > 4
  end

  def status_to_s
     'Competitive release process in progress with active tickets'
  end

  def to_s
    string = "Query ID: #{@query_id}\n" +
      "Domain name: #{@fqdn}\n" +
      "Response status: #{@status} (#{status_to_s})\n" +
      "Release Process:\n" +
      "  Start date: #{@release_process_begin.strftime('%Y-%m-%d %H:%M')}\n" +
      "  End date: #{@release_process_end.strftime('%Y-%m-%d %H:%M')}\n" + 
      "  Accepting new tickets until: #{@accepting_new_tickets_until.strftime('%Y-%m-%d %H:%M')}"
    string += "\nTickets: ##{@tickets.join ', #'}" unless @tickets.empty?
    string
  end
end


module Error
  attr_reader :fqdn, :msg
  def parse(data)
    @data_lines = data.split "\r\n"
    parse_fqdn @data_lines[2] if @data_lines.size > 2

    @msg = @data_lines[3] if @data_lines.size > 3
  end

  def status_to_s
    'Error'
  end

  def to_s
    string = "Query ID: #{@query_id}\n" +
      "Domain name: #{@fqdn}\n" +
      "Response status: #{@status} (#{status_to_s})\n" + 
      "Error: #{@msg}\n" 
    string

  end
end

class Response
  attr_reader :status, :query_id, :cookie, :fqdn, :fqdn_ace

  def initialize(data)
    @data_lines = data.split "\r\n"
    parse_header data

    case @status
    when -1 then extend Undefined
    when 0, 5 then extend AvailableOrWaiting
    when 1 then extend AvailableWithTickets
    when 2 then extend Registered
    when 3 then extend Unavailable
    when 4 then extend Invalid
    when 6, 7 then extend ReleaseInProgress
    when 8 then extend Error
    when 9 then extend CompetitiveReleaseProcess
    end

    parse data
  end

  def parse_header(data)
    data_info = []
    if @data_lines[0][0].chr == '%'
      data_info = @data_lines[1].split if @data_lines.size > 1
    else
      data_info = @data_lines[0].split
    end

    @status = -1
    @cookie = ''
    @query_id = 0

    if data_info[0] == 'CK'
      @cookie = data_info[1] if data_info.size >= 2
      @query_id = data_info[2].to_i if data_info.size >= 3

    elsif data_info[0] == 'ST'
      @status = data_info[1].to_i if data_info.size >= 2
      @query_id = data_info[2].to_i if data_info.size >= 3
    end
  end

  def parse_fqdn(data_line)
    data_info = data_line.split '|'

    @fqdn = ''
    @fqdn_ace = ''

    if data_info.size == 1
      @fqdn = data_info[0]

    elsif data_info.size == 2
      @fqdn_ace = data_info[0]
      @fqdn = data_info[1]
    end
  end
end

#############################################
##                 Client                  ##
#############################################

class Client
  attr_writer :address

  def initialize(address, port, debug)
    @address = address
    @port = port
    @debug = debug
  end

  def ask(query)
    if @debug
      puts '#######################'
      puts "Sending query to #{@address}[#{@port}]"
      puts "Query: #{query}"
    end

    sock = UDPSocket.open
    sock.connect(@address, @port)
    sock.send(query.to_s, 0)
    response = sock.recvfrom(MAX_UDP_SIZE)[0]

    if @debug
      puts "Response: #{response}"
      puts '-----------------------'
    end

    yield response
  end
end

#############################################
##                IsAvail                  ##
#############################################

class IsAvail
  def self.check(query, address, port, cookie_file, debug)
    response = nil

    client = Client.new(address, port, debug)
    client.ask query do |data|
      response = Response.new data

      raise 'Query ID inconsistence' if response.query_id != query.query_id

      if response.cookie != ''
        File.open(cookie_file, "w") { |f| f << response.cookie }
        query.cookie = response.cookie
        response = check(query, address, port, cookie_file, debug)
      end
    end

    response
  end
end
